// Copyright (C) 2006 Erik Dahlberg
// 
// This file is part of LSystem3d.
// 
// LSystem3D is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// LSystem3D is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with LSystem3D; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA




#ifndef MODEL_H
#define MODEL_H

#include <vector>

#include <GL/gl.h>

#include "color.h"

using namespace std;



/**
 * The L-system model
 */
class Model
{
public:
    /**
     * Constructor
     */
    Model();

    /**
     * Destructor
     */
    ~Model();
    
    /**
     * Create a segment
     * @param x1 x start point
     * @param y1 y start point
     * @param z1 z start point
     * @param x2 x end point
     * @param y2 y end point
     * @param z2 z end point
     */
    void segment(double x1, double y1, double z1, double x2, double y2, double z2);
    
    /**
     * Specify a vertex
     * @param x x-coordinate
     * @param y y-coordinate
     * @param z z-coordinate
     */
    void vertex(double x, double y, double z);
    
    /**
     * Specify a normal
     * @param x x-coordinate
     * @param y y-coordinate
     * @param z z-coordinate
     */
    void normal(double x, double y, double z);
    
    /**
     * Render to screen
     */
    void draw();
    
    /**
     * Clear the model
     */
    void clear();
    
    /**
     * Begin a modelling session
     */
    void begin();
    
    /**
     * End a modelling session
     */
    void end();
    
    /**
     * Begin creation of a filled polygon
     */
    void fillBegin();
    
    /**
     * End creation of a filled polygon
     */
    void fillEnd();
    
    /**
     * Increment current index to color table
     */
    void nextColor();
    
    /**
     * Decrement current index to color table
     */
    void prevColor();
    
    /**
     * Decrement diameter of segment
     */
    void decrementDiameter();
    
    /**
     * Set current color index
     * @param index the color index
     */
    void setColorIndex(int index);
    
    /**
     * Set current diameter of segment
     * @param diameter the diameter
     */
    void setDiameter(double diameter);
    
    /**
     * Set diameter factor
     * @param diameter the diameter factor
     */
    void setDiameterFactor(double diameterFactor);
    
    /**
     * Get current color index
     * @return color index
     */
    int getColorIndex();
    
    /**
     * Get current diameter of segment
     * @return diameter of segment
     */
    double getDiameter();
    
protected:
    
    GLuint _displayList;    // All drawing operations
    
    vector<Color> _colorTable;  // Color table
    int _colorIndex;            // Current index to color table
    double _diameter;           // Current diameter of segment
    double _diameterFactor;     // Diameter factor // TODO: "Diameter factor"?
};



#endif
